/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * gclc:net.bigeon.gclc.command.ParametrizedCommand.java
 * Created on: Dec 24, 2014
 */
package net.bigeon.gclc.command;

/*-
 * #%L
 * Generic Command Ligne console
 * %%
 * Copyright (C) 2014 - 2018 bigeon.fr
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */
import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import net.bigeon.gclc.exception.CommandParsingException;
import net.bigeon.gclc.exception.InvalidParameterException;
import net.bigeon.gclc.manager.ConsoleInput;
import net.bigeon.gclc.manager.EmptyInput;

/** An object to handle standardized command parameters.
 *
 * @author Emmanuel BIGEON */
public final class ParametrizedCommandData {

    /** The boolean parameters mandatory status. */
    private final Set<String> boolParams = new HashSet<>();

    /** The string parameters mandatory status. */
    private final Map<String, Boolean> stringParams = new ConcurrentHashMap<>();
    /** The parameters mandatory status. */
    private final Map<String, Boolean> params       = new ConcurrentHashMap<>();
    /** The restriction of provided parameters on execution to declared paramters in
     * the status maps. */
    private final boolean              strict;

    /** The data for the parametrized command. */
    public ParametrizedCommandData() {
        this(true);
    }

    /** The data for a parametrized command.
     *
     * @param strict if the command was refusing unrecognized parameters */
    public ParametrizedCommandData(final boolean strict) {
        this.strict = strict;
    }

    /** Add a boolean parameter to defined parmaters.
     *
     * @param flag the boolean flag
     * @throws InvalidParameterException if the parameter is already defined as a
     *         string parameter */
    public void addBooleanParameter(final String flag) {
        if (params.containsKey(flag) && stringParams.containsKey(flag)) {
            throw new InvalidParameterException("Parameter is already defined as string"); //$NON-NLS-1$
        }
        boolParams.add(flag);
        params.put(flag, Boolean.FALSE);
    }

    /** Add a string parameter to defined parmaters.
     *
     * @param flag the parameter flag
     * @param needed if the parameter's absence should cause an exception
     * @throws InvalidParameterException if the parameter is already defined as a
     *         boolean parameter */
    public void addStringParameter(final String flag, final boolean needed) {
        if (params.containsKey(flag)) {
            checkParam(flag, needed);
            return;
        }
        stringParams.put(flag, Boolean.valueOf(needed));
        params.put(flag, Boolean.valueOf(needed));
    }

    /** Check a parameter.
     *
     * @param param the string parameter
     * @param needed if the parameter is needed
     * @throws InvalidParameterException if the new definition is invalid */
    private void checkParam(final String param, final boolean needed) {
        final Boolean val = stringParams.computeIfPresent(param,
                (final String k, final Boolean v) -> {
                    final Boolean need = Boolean.valueOf(needed || v.booleanValue());
            params.put(param, need);
                    return need;
                });
        if (val == null) {
            throw new InvalidParameterException(
                    "Parameter is already defined as boolean"); //$NON-NLS-1$
        }
    }

    /** Retrieve the boolean parameters (aka flags).
     *
     * @return the set of boolean parameters */
    public Set<String> getBooleanParameters() {
        return Collections.unmodifiableSet(boolParams);
    }

    /** Retrieve the parameter names.
     *
     * @return the stringParams */
    public Set<String> getParameters() {
        return params.keySet();
    }

    /** Get the parameters from an input.
     *
     * @param input the input
     * @param args the command arguments
     * @return the command object
     * @throws IOException if the command could not be filled. */
    public CommandParameters getParameters(final ConsoleInput input, final String... args)
            throws IOException {
        final CommandParameters parameters = new CommandParameters(boolParams,
                stringParams.keySet(), strict);
        try {
            parameters.parseArgs(args);
        } catch (final CommandParsingException e) {
            throw new IOException(e);
        }
        final List<String> toProvide = new ArrayList<>();
        for (final Entry<String, Boolean> string : params.entrySet()) {
            if (string.getValue().booleanValue()
                    && parameters.get(string.getKey()) == null) {
                if (input == null || input == EmptyInput.INSTANCE) {
                    throw new IOException();
                }
                toProvide.add(string.getKey());
            }
        }
        // for each needed parameters that is missing, prompt the user.
        fillParameters(input, toProvide, parameters);
        return parameters;
    }

    /** Get the string parameters names.
     *
     * @return the stringParams */
    public Set<String> getStringParameters() {
        return stringParams.keySet();
    }

    /** Test if a parameter is needed.
     *
     * @param param the parameter name
     * @return if the parameter is needed */
    public boolean isNeeded(final String param) {
        final Boolean val = params.get(param);
        return val != null && val.booleanValue();
    }

    /** If the command refuse unrecognized parameters.
     *
     * @return the strict */
    public boolean isStrict() {
        return strict;
    }

    /** Fill the undefined parameters.
     * <p>
     * This method prompts the user to fill the needed parameters.
     *
     * @param input the input to prompt through
     * @param parameters the parameter list to complete
     * @param toProvide the parameters to ask for
     * @throws IOException if the manager was closed */
    private static void fillParameters(final ConsoleInput input,
            final List<String> toProvide, final CommandParameters parameters)
            throws IOException {
        for (final String string : toProvide) {
            String value;
            value = input.prompt(MessageFormat.format("value of {0}? ", string)); //$NON-NLS-1$
            while (value.isEmpty()) {
                value = input
                        .prompt(MessageFormat.format("value of {0}? (cannot be empty) ", //$NON-NLS-1$
                                string));
            }
            parameters.set(string, value);
        }
    }
}
